import uuid
from functools import reduce
from sys import stderr
from typing import Dict

error_key = '__error__'
validation_errors = []

valid_content_types = {
    'text',
    'char',
    'integer',
    'float',
    'uri',
    'url',
    'bool',
    'existing_lit_predicate',
    'new_or_existing_lit_predicate',
    'single_choice',
    'multiple_choice'
}

def validate_api():
    try:
        log_validation_info('Starting validation')
        try:
            from plugin import callbacks
        except ImportError as e:
            log_validation_error(f"Could not find plugin code (Detail: import 'callbacks' from package 'plugin' failed). {e}")
            return False

        succeeded = _validate_metadata(callbacks)
        succeeded = _validate_configuration_options(callbacks) and succeeded
        succeeded = _validate_actions(callbacks) and succeeded
        if succeeded:
            log_validation_info('Starting example validations')
            _validate_input_output(callbacks)
        else:
            log_validation_info('Skipping unit tests due to other errors')
    except Exception as e:
        log_validation_error(f"Unexpected error while validating plugin code: {e}.")
    log_validation_info('Validation done')
    return succeeded


def _validate_metadata(callbacks):
    metadata = _get_function_value(callbacks, 'meta_data')
    if metadata is None:
        return False
    if not isinstance(metadata, dict):
        log_validation_error(f"Plugin code object 'meta_data' should return a dictionary.")
        return False
    succeeded = _check_dict_elements(metadata, "the return value of function metadata",
                       (
                           ('id', check_type(str, f"should be a string"), False),
                           ('name', check_type(str, f"should be a string"), False),
                           ('description', check_type(str, f"should be a string"), False),
                           ('type-id', lambda x: raise_if_not(x == 'batch-transform',
                                                              "can only be equal to 'batch-transform'."), True),
                           ('api-version', lambda x: raise_if_not(x == '1.0.0', "can only be equal to '1.0.0'."), True),
                       ))
    if metadata:
        log_validation_info(f"The return value of function metadata contains unknown keys: they will be ignored. "
                            f"Keys: {','.join(metadata.keys())}")
    return succeeded


def _validate_configuration_options(callbacks):
    configuration_options = _get_function_value(callbacks, 'configuration_options')
    if configuration_options is None:
        return False
    if not isinstance(configuration_options, list):
        log_validation_error(f"Plugin code object 'configuration_options' should return a list.")
        return False
    return reduce(lambda sum, el: sum and _validate_option(el[1], el[0]), enumerate(configuration_options), True)


def _validate_actions(callbacks):
    succeeded = _validate_function(callbacks, "initialize_session")
    succeeded = succeeded and _validate_function(callbacks, "preferred_batch_size")
    succeeded = succeeded and _validate_function(callbacks, "process_resources")
    succeeded = succeeded and _validate_function(callbacks, "finalize_session")
    return  succeeded

def _validate_input_output(callbacks):
    get_example_function = _get_function(callbacks, 'validate_examples', optional=True)
    if not get_example_function:
        return True
    examples = get_example_function()
    options = _get_function_value(callbacks, 'configuration_options')
    input_ids = set(opt['id'] for opt in options
                    if opt['content_type'] == 'existing_lit_predicate')
    output_ids = set(opt['id'] for opt in options
                    if opt['content_type'] == 'new_or_existing_lit_predicate')
    try:
        for i, example in enumerate(examples):
            try:
                validate_example_(callbacks, example, input_ids, output_ids)
            except ValidationError as e:
                log_validation_warning(f'Error while evaluating session {i}: {e}')
            except Exception as e:
                log_validation_warning(f'Unexpected error while evaluating session {i}: {e}.')
    except:
        log_validation_warning(f"The return value of the function validate_examples should be iterable.")
    get_example_cleanup_function = _get_function(callbacks, 'validate_examples_cleanup', optional=True)
    if get_example_cleanup_function:
        get_example_cleanup_function()


def validate_example_(callbacks, example_session, input_ids, output_ids):
    if not isinstance(example_session, tuple) or len(example_session) != 2:
        raise ValidationError(f"The return value of the function validate_examples should be list of tuples,"
                             f" containing option values (dict) and batch (list of tuples).")
    option_values = example_session[0]
    if not isinstance(option_values, dict):
        raise ValidationError(f"the first element in the tuple (options) should be a dict.")
    if not isinstance(example_session[1], list) or not all(isinstance(el, tuple) and len(el)==2 for el in example_session[1]):
        raise ValidationError(f"the batch should be a list of tuples containing example input and output dicts.")
    batch_input = []
    expected_batch_output = []
    for nr, inp_out in enumerate(example_session[1]):
        input, output = inp_out
        if not isinstance(input, dict) or not isinstance(output, dict):
            if not isinstance(input, dict):
                raise ValidationError(f"the first element in the tuple number {nr+1} (input) should be a dict.")
            if not isinstance(output, dict):
                raise ValidationError(f"the second element in the tuple number {nr+1} (output) should be a dict.")
        if set(input.keys()) != input_ids:
            raise ValidationError(f"input keys should correspond to ids of 'existing_lit_predicate' options: "
                                  f"expected {input_ids}, received {input.keys()}.")
        # if error is present, the rest may be present but doesn't matter
        if '__error__' not in output.keys() and set(output.keys()) != output_ids:
            raise ValidationError(f"output keys should correspond to ids of 'new_or_existing_lit_predicate' options: "
                                  f"expected {output_ids}, received {output.keys()}.")
        if not values_are_list_of_strings(input):
            raise ValidationError(f"input values should be a list of strings.")
        if not values_are_list_of_strings(output):
            raise ValidationError(f"output values should be a list of strings.")
        batch_input.append(input)
        expected_batch_output.append(output)
    # do the test
    session_id = str(uuid.uuid4())
    _get_function_value(callbacks, "initialize_session", (session_id, option_values))
    try:
        generated_batch_output = _get_function_value(callbacks, "process_resources", (session_id, option_values, batch_input))
    except Exception as e:
        raise ValidationError(f"Unexpected error encountered: {e}")
    _get_function_value(callbacks, "finalize_session", (session_id, option_values))
    if len(expected_batch_output) != len(generated_batch_output):
        raise ValidationError(f"only {len(generated_batch_output)} output values were generated, while {len(expected_batch_output)} expected.")
    success = True
    for generated_output, expected_output in zip(generated_batch_output, expected_batch_output):
        if generated_output != expected_output:
            log_validation_warning(f"expected output {expected_output} and received {generated_output}.")
            success = False
    return success


def values_are_list_of_strings(dct):
    return all(isinstance(value, list) and all(isinstance(el, str) for el in value) for key,value in dct.items() if key!=error_key)


def _validate_function(callbacks, function_name):
    func = _get_function(callbacks , function_name)
    return func is not None


def _validate_option(option, nr):
    if not isinstance(option, dict):
        log_validation_error(f"Each element returned by function configuration_options should be a dictionary."
                             f"Element {nr} is a {type(option).__name__}.")
        return False
    # check the interactions
    dict_status = True
    if 'default' in option and option.get('content_type') in ('existing_lit_predicate', 'new_or_existing_lit_predicate'):
        log_validation_error(f"You cannot specify a default for input or output predicate.")
        dict_status = False
    if 'choices' in option and option.get('content_type') not in (None, 'single_choice', 'multiple_choice'):
        log_validation_error(f"The key 'choices' can only be used with content type 'single_choice' or 'multiple_choice'.")
        dict_status = False
    if 'choices' not in option and option.get('content_type') in ('single_choice', 'multiple_choice'):
        log_validation_error(f"The key 'choices' must be present with content type 'single_choice' or 'multiple_choice'.")
        dict_status = False
    dict_status = dict_status and\
                  _check_dict_elements(option, "each element returned by function configuration_options",
                         (
                             ('id', check_type(str, f"should be a string"), False),
                             ('name', check_type(str, f"should be a string"), False),
                             ('description', check_type(str, f"should be a string"), False),
                             ('default', None, True),
                             ('section', check_type(str, f"should be a string"), True),
                             ('choices', check_list_of_dicts("The value of key choices ", (
                                 ('id', check_type(str, f"should be a string"), False),
                                 ('name', check_type(str, f"should be a string"), False),
                                 ('description', check_type(str, f"should be a string"), True)
                             )), True),
                             ('content_type', lambda x: raise_if_not(x in valid_content_types,
                                       f" should be one of {', '.join(valid_content_types)}."), False)
                         ))
    return dict_status


def _get_function(callbacks, function_name, optional=False):
    if not hasattr(callbacks, function_name):
        if not optional:
            log_validation_error(f"Plugin code does not contain function '{function_name}'.")
        return None
    metadata_func = getattr(callbacks,function_name)
    if not callable(metadata_func):
        log_validation_error(f"Plugin code object '{function_name}' should be a function (callable).")
        return None
    return metadata_func


def _get_function_value(callbacks, function_name, args=None):
    metadata_func = _get_function(callbacks, function_name)
    if not metadata_func:
        return None
    if args:
        return metadata_func(*args)
    else:
        return metadata_func()


def _check_dict_elements(dct: Dict, description, expected_elements):
    return reduce(lambda sum, el: _check_dict_has_element(dct,
                                                          el[0],
                                                          description,
                                                          value_validator=el[1],
                                                          optional=el[2])
                                  and sum,  #not sum and becaue lazy evaluation!
                  expected_elements,
                  True)


class ValidationError(Exception):
    pass


def _check_dict_has_element(dct: Dict, key: str, description: str, value_validator=None,
                            optional=False) -> bool:
    """Note: this function removes the key which was investigated.
    This allows to mention unknown keys without repeating the list"""
    if key not in dct:
        if optional:
            return True
        else:
            log_validation_error(f"{description.capitalize()} should contain key '{key}'. ")
            return False
    if not dct[key]:
        log_validation_error(f"The value of key '{key}' in {description} should not be empty.")
        del dct[key]
        return False
    if value_validator:
        try:
            value_validator(dct[key])
        except ValidationError as e:
            if str(e):
                log_validation_error(f"The value of key '{key}' in {description} {e}")
            return False
    del dct[key]
    return True


def check_list_of_dicts(description, expected_elements):
    return  lambda x: raise_if_not(reduce(lambda sum, el: _check_dict_elements(el, description, expected_elements) and sum, x, True),
                                   '')


def check_type(tp, message):
    return lambda x: raise_if_not(isinstance(x, tp), message)


def raise_if_not(condition: bool, message: str):
    if not condition:
        raise ValidationError(message)


def log_validation_info(message):
    print(f"INFO\t{message}")


def log_validation_warning(message):
    print(f"WARNING\t{message}")


def log_validation_error(message):
    validation_errors.append(message)
    print(f"ERROR\t{message}")

if __name__ == '__main__':
    validate_api()
    if not validation_errors:
        print('Validation OK')
    else:
        print('Validation errors found')
        exit(1)